from typing import Any, Optional, Type

from fastapi import Query
from sqlalchemy import text
from assimilator.alchemy.database import AlchemyRepository

from velait.common.services.exceptions import PaginationLimitsError
from velait.common.services.pagination import get_offset_limits
from velait.common.services.search import Search, SearchError, SearchOperator


class AlchemyRepositorySearch(Search):
    def __init__(
        self,
        model: Type['BaseModel'],
        repository: AlchemyRepository,
        search_: Optional[str] = None,
        query: str = Query(default=None),
        page: Optional[Any] = None,
        ordering: Optional[str] = None,
    ):
        self.repository = repository
        super(AlchemyRepositorySearch, self).__init__(
            search=search_,
            query=query,
            page=page,
            ordering=ordering,
            model=model,
        )

    def _parse_ordering(self):
        alchemy_ordering = list(self._ordering)

        for i in range(len(alchemy_ordering)):
            if alchemy_ordering[i].startswith('-'):
                alchemy_ordering[i] = f"{alchemy_ordering[i][1:]} desc"

            alchemy_ordering[i] = text(alchemy_ordering[i])

        return alchemy_ordering

    def search(self):
        try:
            filter_spec = self.repository.specifications.filter(*self.parse_query_filters())

            offset, limit = get_offset_limits(offset=self._page['offset'], page_size=self._page['size'])
            pagination_spec = self.repository.specifications.paginate(offset=offset, limit=limit)

            if self._ordering:
                return self.repository.filter(
                    filter_spec,
                    pagination_spec,
                    self.repository.specifications.order(*self._parse_ordering()),
                    lazy=True,
                )

            return self.repository.filter(filter_spec, pagination_spec, lazy=True)

        except PaginationLimitsError:
            raise SearchError(
                name='page',
                description="'page' parameter contains invalid values",
            )
        except Exception:
            raise SearchError(
                name="search",
                description="Search could not be conducted",
            )

    def _parse_operator(self, field_name: str, operator: str, field_value: str):
        """
        Creates an expression object if all input operators are valid.
        If they are not, raises op an exception
        """

        if operator == "equals":
            return getattr(self.model, field_name) == field_value
        elif operator == "lessOrEqual":
            return getattr(self.model, field_name) <= field_value
        elif operator == "greaterOrEqual":
            return getattr(self.model, field_name) >= field_value
        elif operator == "greater":
            return getattr(self.model, field_name) > field_value
        elif operator == "less":
            return getattr(self.model, field_name) < field_value
        elif operator == "contains":
            return getattr(self.model, field_name).in_(field_value)
        else:
            raise SearchError(
                name="query",
                description=f"Operation '{operator}' is unknown",
            )


__all__ = [
    'SearchError',
    'SearchOperator',
    'AlchemyRepositorySearch',
]



from typing import Iterable, Union

from fastapi import Request
from dependency_injector.wiring import inject
from starlette.datastructures import URL
from assimilator.core.database import LazyCommand, BaseRepository

from velait.velait_fastapi.api.pagination.schemas import Page, PageInfo


def get_next_page(url: URL, current_page: int, last_page: int):
    if current_page == last_page:
        return None

    return str(url.include_query_params(page=current_page + 1))


def get_prev_page(url: URL, current_page: int):
    if current_page == 0:
        url.remove_query_params('page')
        return str(url)

    return str(url.include_query_params(page=current_page - 1))


@inject
def paginate(
    request: Request,
    page_size: int,
    data: Union[LazyCommand, Iterable],
    repository: BaseRepository,
) -> Page:
    items = data() if isinstance(data, LazyCommand) else data
    total_count = repository.count()
    last_page = total_count // page_size

    return Page(
        results=items,
        pagination=PageInfo(
            total_records=total_count,
            total_pages=last_page,
            first=str(request.url.remove_query_params('page')),
            last=str(request.url.include_query_params(page=last_page)),
            next=get_next_page(
                request.url,
                last_page=last_page,
                current_page=request.path_params.get('page', 0)
            ),
            previous=get_prev_page(request.url, current_page=request.path_params.get('page', 0)),
        ),
    )


__all__ = [
    'paginate',
]



import uuid

from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.sql.sqltypes import TIMESTAMP
from sqlalchemy import Column, String, func


_base, _named = None, None


def create_base_models(Base):
    global _base, _named

    if _base is not None:
        return _base, _named

    class BaseModel(Base):
        __abstract__ = True
        id = Column(UUID(as_uuid=True), default=uuid.uuid4, primary_key=True, unique=True)

        created_at = Column(TIMESTAMP, server_default=func.now())
        updated_at = Column(TIMESTAMP, server_default=func.now(), onupdate=func.now())
        created_by_id = Column(UUID(as_uuid=True), default=uuid.uuid4)
        updated_by_id = Column(UUID(as_uuid=True), default=uuid.uuid4)

        queryable_fields = ("id", "created_at", "updated_at", "created_by_id", "updated_by_id")
        orderable_fields = ("id", "created_at", "updated_at", "created_by_id", "updated_by_id")
        searchable_fields = ("id", "created_at", "updated_at", "created_by_id", "updated_by_id")
        searched_values = ("id", "created_at", "updated_at", "created_by_id", "updated_by_id")

    class NameModel(BaseModel):
        __abstract__ = True
        name_eng = Column(String, unique=True)
        name_kaz = Column(String, unique=True)
        name_rus = Column(String, unique=True)

        queryable_fields = BaseModel.queryable_fields + ('name_eng', 'name_kaz', 'name_rus')
        orderable_fields = BaseModel.orderable_fields + ('name_eng', 'name_kaz', 'name_rus')
        searchable_fields = BaseModel.searchable_fields + ('name_eng', 'name_kaz', 'name_rus')
        searched_values = BaseModel.searched_values + ('name_eng', 'name_kaz', 'name_rus')

    _base, _named = BaseModel, NameModel
    return _base, _named


__all__ = [
    'create_base_models',
]



from typing import Optional, TypeVar, Generic, List

from pydantic.generics import GenericModel

from velait.velait_fastapi.api.schemas import BaseSchema


class PageInfo(BaseSchema):
    total_records: int
    total_pages: int
    first: str
    last: str
    next: Optional[str]
    previous: Optional[str]


T = TypeVar("T")


class Page(GenericModel, Generic[T]):
    results: List[T]
    pagination: PageInfo


__all__ = ['Page', 'PageInfo']



from django.db.models import QuerySet
from django.utils.translation import gettext_lazy as _

from velait.common.exceptions import AlreadyDeletedError
from velait.velait_django.main.models import BaseModel
from velait.velait_django.main.services.decorators import (
    only_decorator,
    get_related_decorator,
    ordering_decorator,
    pagination_decorator,
    first_object_decorator,
)


def get_object(objects, **kwargs):
    return filter_objects(objects, **kwargs, first=True)


@first_object_decorator
@only_decorator
@pagination_decorator
@ordering_decorator
@get_related_decorator
def filter_objects(objects, **kwargs) -> QuerySet:
    return objects.filter(**kwargs)


@first_object_decorator
@only_decorator
@pagination_decorator
@ordering_decorator
@get_related_decorator
def all_objects(objects):
    return filter_objects(objects)


def empty_objects(objects) -> QuerySet:
    return objects.none()


def object_exists(objects, **kwargs) -> bool:
    return filter_objects(objects, **kwargs).exists()


def delete_object(model: BaseModel):
    if model.is_deleted:
        raise AlreadyDeletedError(name="obj", description=_("Уже удалено"))

    model.is_deleted = True
    model.save(update_fields=['is_deleted'])


__all__ = [
    'get_object',
    'filter_objects',
    'all_objects',
    'empty_objects',
    'object_exists',
]



from velait.velait_django.velait_users.views import user_logout_view, user_login_view, user_authentication_view
from django.urls import path

app_name = "apps.velait_users"

urlpatterns = [
    path("logout/", user_logout_view, name="logout"),
    path("login/", user_login_view, name="login"),
    path('authenticate/', user_authentication_view, name="authenticate"),
]



import json

import requests
from django.conf import settings
from django.contrib.auth import get_user_model

from velait.velait_django.velait_users.services import authorize_user


User = get_user_model()


class AuthenticationMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def get_user_data(self, access_token):
        try:
            response = requests.get(
                settings.MIO_SSO_USER_INFO_URL,
                headers={"Authorization": f"Bearer {access_token}"},
            )
            if not response.ok:
                return None

            user_data = response.json()
            return User(
                business=user_data.get('business'),
                id=user_data.get('sub'),
                uuid=user_data.get('uuid'),
                username=user_data.get('preferred_username'),
                email=user_data.get('email'),
                first_name=user_data.get('given_name'),
                last_name=user_data.get('family_name'),
                is_staff=user_data.get('is_staff'),
                #   "groups": response_data['groups'],
            )
        except (json.JSONDecodeError, requests.RequestException):
            return None

    def _refresh_user(self, request):
        try:
            access_token, refresh_token, expiration_time = authorize_user(
                grant_type="refresh_token",
                auth_data=(
                    ('refresh_token', request.COOKIES.get('refresh_token')),
                ),
            )
        except ValueError:
            return self.get_response(request)

        request.user = self.get_user_data(access_token)

        response = self.get_response(request)
        response.set_cookie(
            key='access_token',
            value=access_token,
            expires=expiration_time,
            httponly=True,
            secure=True,
        )
        response.set_cookie(
            key='refresh_token',
            value=refresh_token,
            expires=expiration_time,
            httponly=True,
            secure=True,
        )
        return response

    def __call__(self, request):
        if not request.COOKIES.get('access_token') and not request.COOKIES.get('access_token'):
            return self.get_response(request)

        user_data = self.get_user_data(request.COOKIES.get('access_token'))
        if user_data is None:
            return self._refresh_user(request)

        request.user = user_data
        return self.get_response(request)



from datetime import datetime

from velait.velait_fastapi.api.users.exceptions import NotAuthorizedError
from velait.velait_fastapi.api.users.schemas import UserSchema
from velait.velait_fastapi.connections.exceptions import RequestError
from velait.velait_fastapi.connections.http import HTTPClient


class SSOService:
    def __init__(self, sso_client: HTTPClient):
        self.sso_client = sso_client

    def _create_authorization_headers(self, token: str):
        return {"Authorization": token}

    async def get_user_info(self, token: str):
        try:
            if not token:
                raise NotAuthorizedError()

            user_info = await self.sso_client.request(
                "/o/userinfo/", "GET", headers=self._create_authorization_headers(token))
        except RequestError:
            raise NotAuthorizedError()

        return UserSchema(
            id=user_info['uuid'],
            username=user_info['preferred_username'],
            email=user_info['email'],
            permissions=user_info['groups'],
            created_at=datetime.now(),
        )


__all__ = ['SSOService']



from fastapi import Request
from starlette.middleware.base import BaseHTTPMiddleware
from assimilator.core.database.exceptions import InvalidQueryError, NotFoundError

from velait.common.services.search import SearchError
from velait.velait_fastapi.api.responses import APIResponse, ResponseErrorItem
from velait.velait_fastapi.api.users.exceptions import NotAuthorizedError
from velait.velait_fastapi.api.users.permissions import NoPermissionError
from velait.common.exceptions import AlreadyDeletedError


class ExceptionHandlerMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        try:
            response = await call_next(request)
        except NotFoundError:
            return APIResponse(status_code=404, errors=[
                ResponseErrorItem(name="obj", description="Not found")
            ])
        except InvalidQueryError:
            return APIResponse(status_code=400, errors=[
                ResponseErrorItem(name="obj", description="Invalid request data")
            ])
        except (NotAuthorizedError, NoPermissionError, SearchError, AlreadyDeletedError) as exc:
            return APIResponse(status_code=exc.status_code, errors=[
                ResponseErrorItem(name=exc.name, description=exc.description)
            ])

        return response


__all__ = [
    'ExceptionHandlerMiddleware',
]



import json
from datetime import timedelta
from typing import Tuple, Iterable

import requests
from django.conf import settings
from django.utils import timezone

from velait.velait_django.velait_users.exceptions import UserAuthenticationError


def authorize_user(grant_type: str, auth_data: Iterable[Tuple[str, str]]):
    try:
        auth_response = requests.post(
            url=settings.MIO_SSO_TOKEN_URL,
            data=[
                ("grant_type", grant_type),
                ("scope", " ".join(settings.MIO_SSO_SCOPES)),
                ("client_id", settings.MIO_SSO_CLIENT_KEY),
                ("client_secret", settings.MIO_SSO_CLIENT_SECRET),
                *auth_data,
            ],
        )

        if not auth_response.ok:
            raise requests.RequestException()

        json_data = auth_response.json()
        access_token = json_data['access_token']
        refresh_token = json_data['refresh_token']
        expires_in = json_data['expires_in']
    except (json.JSONDecodeError, KeyError, requests.RequestException):
        raise UserAuthenticationError()

    return access_token, refresh_token, timezone.now() + timedelta(seconds=expires_in)



from assimilator.alchemy.database import AlchemyRepository as BaseAlchemyRepository
from velait.common.exceptions import AlreadyDeletedError


class VelaitRepository(BaseAlchemyRepository):
    def delete(self, obj):
        if obj.is_removed:
            raise AlreadyDeletedError()

        obj.is_removed = True


__all__ = [
    'VelaitRepository',
]



from velait.common.exceptions import VelaitError
from velait.velait_fastapi.connections.exceptions import RequestError


class NotAuthorizedError(RequestError):
    name = "user"
    description = "User is not authorized"
    status_code = 403


class NoPermissionError(VelaitError):
    name = "user"
    description = "User does not have enough permissions"
    status_code = 403


__all__ = ['NotAuthorizedError', 'NoPermissionError']



from typing import Type

from rest_framework.views import APIView
from rest_framework.serializers import ModelSerializer

from velait.velait_django.main.models import BaseModel
from velait.velait_django.main.api.responses import APIResponse
from velait.velait_django.main.services.search import SearchError, DjangoSearch
from velait.velait_django.main.api.pagination import VelaitPagination


class SearchView(APIView):
    model: Type[BaseModel] = None
    serializer_class: Type[ModelSerializer] = None
    search_class: Type[DjangoSearch] = DjangoSearch

    def __init__(self, *args, **kwargs):
        super(SearchView, self).__init__(*args, **kwargs)

        if self.model is None or self.serializer_class is None:
            raise NotImplementedError("Model or Serializer were not supplied to the SearchView")

    def get_search_object(self):
        return self.search_class(
            search_=self.request.GET.get('search'),
            query=self.request.GET.get('query'),
            ordering=self.request.GET.get('ordering'),
            page=self.request.GET.get('page'),
            model=self.model,
            paginate=False,
        )

    def get(self, request, *args, **kwargs):
        try:
            search = self.get_search_object()
            paginator = VelaitPagination()
            queryset = paginator.paginate_queryset(queryset=search.search(), request=request, view=self)
            return paginator.get_paginated_response(self.serializer_class(instance=queryset, many=True).data)

        except SearchError as exc:
            return APIResponse(errors=[exc], status=400)




class VelaitError(Exception):
    name: str
    description: str
    status_code: int = 400

    def __init__(self, name: str = None, description: str = None, status_code: int = None, *args):
        super(VelaitError, self).__init__(*args)
        if name is not None:
            self.name = name

        if description is not None:
            self.description = description

        if status_code is not None:
            self.status_code = status_code


class AlreadyDeletedError(VelaitError):
    name = "obj"
    description = "Already deleted"
    status_code = 404


__all__ = ['AlreadyDeletedError', 'VelaitError']



from rest_framework.serializers import ModelSerializer

from velait.velait_django.main.models import BaseModel


class BaseSerializer(ModelSerializer):
    class Meta:
        model = BaseModel
        fields = (
            "uuid",
            "is_deleted",
            "created_at",
            "created_by_id",
            "updated_at",
            "updated_by_id",
        )
        read_only_fields = (
            "uuid",
            "is_deleted",
            "created_at",
            "created_by_id",
            "updated_at",
            "updated_by_id",
        )
        abstract = True


__all__ = ['BaseSerializer']



from djangorestframework_camel_case.parser import CamelCaseFormParser
from djangorestframework_camel_case.parser import CamelCaseMultiPartParser
from djangorestframework_camel_case.parser import CamelCaseJSONParser


class VelaitFormParser(CamelCaseFormParser):
    pass


class VelaitMultiPartParser(CamelCaseMultiPartParser):
    pass


class VelaitAPIParser(CamelCaseJSONParser):
    pass


__all__ = [
    'VelaitFormParser',
    'VelaitMultiPartParser',
    'VelaitAPIParser',
]



from rest_framework.views import exception_handler

from velait.common.exceptions import VelaitError
from velait.velait_django.main.api.responses import APIResponse


def custom_exception_handler(exc, context):
    response = exception_handler(exc, context)

    if response is None:
        raise exc
    elif isinstance(response.data, dict):
        errors = [
            VelaitError(name=error_name, description=error_description)
            for error_name, error_description in response.data.items()
        ]
    else:
        errors = response.data

    return APIResponse(
        data=None,
        status=response.status_code,
        template_name=response.template_name,
        headers=response.headers,
        exception=response.exception,
        content_type=response.content_type,
        errors=errors,
    )



from typing import Optional
from datetime import datetime
from uuid import UUID

from pydantic import BaseModel


def to_camel_case(value: str):
    words = value.split('_')
    return words[0] + ''.join(word.capitalize() for word in words[1:])


class BaseSchema(BaseModel):
    class Config:
        orm_mode = True
        alias_generator = to_camel_case
        allow_population_by_field_name = True


class ReadBaseSchema(BaseSchema):
    id: UUID
    created_at: datetime
    updated_at: Optional[datetime]
    created_by_id: Optional[UUID]
    updated_by_id: Optional[UUID]


class WriteBaseSchema(BaseSchema):
    class Config:
        orm_mode = True
        alias_generator = to_camel_case
        allow_population_by_field_name = True


__all__ = [
    'BaseSchema',
    'ReadBaseSchema',
    'WriteBaseSchema',
]



from setuptools import setup, find_packages

setup(
    name='velait_api',
    version='0.1.3',
    license='LICENSE.md',
    packages=find_packages(),
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    optional_requires={
        'django': ['django', 'djangorestframework', 'djangorestframework-camel-case'],
        'fastapi': ['fastapi', 'sqlalchemy', 'py-assimilator==0.2.5', 'dependency-injector', 'aiohttp'],
    },
)



import aiohttp
from aiohttp import ClientError

from velait.velait_fastapi.connections.exceptions import RequestError


class HTTPClient:
    def __init__(self, base_url: str):
        self.base_url = base_url

    async def request(self, url, method, headers=None):
        async with aiohttp.ClientSession() as session:
            try:
                async with session.request(method, f"{self.base_url}{url}", headers=headers) as response:
                    if not response.ok:
                        raise RequestError(name="request", description=response.reason, status_code=response.status)

                    return await response.json()
            except ClientError:
                raise RequestError(name="request", description="Unknown request error")


__all__ = [
    'HTTPClient',
    'RequestError',
]



from typing import Iterable

from fastapi.responses import JSONResponse

from velait.velait_fastapi.api.schemas import BaseSchema


class APIResponse(JSONResponse):
    def __init__(
        self,
        content=None,
        status_code=200,
        headers=None,
        errors: Iterable['ResponseErrorItem'] = None,
        **kwargs
    ):
        if errors is not None:
            content = {
                "Errors": [error.dict() for error in errors],
            }

        super(APIResponse, self).__init__(content, status_code=status_code, headers=headers, **kwargs)


class ResponseErrorItem(BaseSchema):
    name: str
    description: str


__all__ = [
    'APIResponse',
    'ResponseErrorItem',
]



from rest_framework.response import Response


class APIResponse(Response):
    def __init__(
        self,
        data=None,
        status=None,
        template_name=None,
        headers=None,
        exception=False,
        content_type=None,
        errors=None,
    ):
        if errors is not None:
            data = {"errors": [{'name': error.name, 'description': error.description} for error in errors]}
        elif not (isinstance(data, dict) and data.get('pagination') is not None):
            data = {"results": data}

        super(APIResponse, self).__init__(
            data=data,
            status=status,
            template_name=template_name,
            headers=headers,
            exception=exception,
            content_type=content_type,
        )



from typing import Iterable

from fastapi import Request

from velait.velait_fastapi.api.users.exceptions import NoPermissionError


class UserGroups:
    ADMIN = "Администратор"
    STAFF = "staff"
    READER = "reader"


class PermissionChecker:
    def __init__(self, permissions: Iterable[str]):
        self.permissions = set(permissions)

    def __call__(self, request: Request = None):
        if not set(request.user.permissions) & set(self.permissions):
            raise NoPermissionError()


__all__ = ['NoPermissionError', 'UserGroups', 'PermissionChecker']



import json

from django.conf import settings
from rest_framework.response import Response
from rest_framework.utils.urls import replace_query_param
from rest_framework.pagination import PageNumberPagination


class VelaitPagination(PageNumberPagination):
    def get_last_page(self):
        return replace_query_param(
            url=self.request.build_absolute_uri(),
            key=self.page_query_param,
            val=self.page.paginator.num_pages,
        )

    def __parse_page_param(self, request):
        try:
            parsed_page = json.loads(request.query_params['page'])

            if not isinstance(parsed_page, dict):
                raise TypeError()
            elif parsed_page.get('size') <= 0 or parsed_page.get('offset') < 0:
                raise ValueError()

            return parsed_page

        except (json.JSONDecodeError, TypeError, ValueError, KeyError):
            return {'offset': 0, 'size': settings.REST_FRAMEWORK['PAGE_SIZE']}

    def get_page_size(self, request):
        page_param = self.__parse_page_param(request)
        if page_param:
            return page_param.get('size')

        return super(VelaitPagination, self).get_page_size(request)

    def get_page_number(self, request, paginator):
        page_param = self.__parse_page_param(request)

        if page_param:
            page = page_param.get('offset') // page_param.get('size')
            if page <= 0:
                return 1
            return page

        return super(VelaitPagination, self).get_page_number(request, paginator)

    def get_paginated_response(self, data):
        return Response({
            "pagination": {
                'totalRecords': self.page.paginator.count,
                'totalPages': self.page.paginator.num_pages,
                'first': self.request.build_absolute_uri(),
                'last': self.get_last_page(),
                'next': self.get_next_link(),
                'previous': self.get_previous_link(),
            },
            "results": data,
        })



from urllib.parse import urlencode

from django.conf import settings
from django.shortcuts import redirect
from django.core.exceptions import BadRequest
from django.http import HttpResponseRedirect
from django.utils.translation import gettext_lazy as _

from velait.velait_django.velait_users.services import authorize_user
from velait.velait_django.main.services.utils import create_random_str


def user_logout_view(request):
    response = HttpResponseRedirect(settings.LOGOUT_REDIRECT_URL)
    response.delete_cookie('access_token')
    response.delete_cookie('refresh_token')
    return response


def user_login_view(request):
    request_args = urlencode({
        "response_type": "code",
        "state": create_random_str(getattr(settings, 'USER_LOGIN_STATE_LEN', 64)),
        "nonce": create_random_str(getattr(settings, 'USER_LOGIN_NONCE_LEN', 64)),
        "scope": " ".join(settings.MIO_SSO_SCOPES),
        "client_id": settings.MIO_SSO_CLIENT_KEY,
        "redirect_uri": settings.MIO_SSO_REDIRECT_URL,
    })

    return redirect(f"{settings.MIO_SSO_AUTHORIZATION_URL}?{request_args}")


def user_authentication_view(request):
    try:
        access_token, refresh_token, expiration_time = authorize_user(
            grant_type="authorization_code",
            auth_data=[('code', request.GET.get('code'))]
        )
    except ValueError:
        raise BadRequest(_("Ошибка авторизации"))

    response = HttpResponseRedirect(settings.LOGIN_REDIRECT_URL)
    response.set_cookie(
        key='access_token',
        value=access_token,
        expires=expiration_time,
        httponly=True,
        secure=True,
    )
    response.set_cookie(
        key='refresh_token',
        value=refresh_token,
        expires=expiration_time,
        httponly=True,
        secure=True,
    )
    return response



from velait.common.services.exceptions import PaginationLimitsError


def get_page_limits(page: int, page_size: int):
    if page_size <= 0 or page < 0:
        raise PaginationLimitsError()

    offset = page * page_size
    return offset, offset + page_size


def get_offset_limits(offset: int, page_size: int):
    if page_size <= 0 or offset < 0:
        raise PaginationLimitsError()

    return offset, page_size


__all__ = ['get_page_limits', 'get_offset_limits']
